package ssw.mj.codegen;

import java.io.IOException;
import java.io.OutputStream;

import ssw.mj.ErrorStrings;
import ssw.mj.Parser;
import ssw.mj.symtab.Tab;

/**
 * MicroJava Code Generator.
 * 
 * @author AW
 */
public class Code {
	// instruction codes
	public static final int
			load        =  1,
			load_n      =  2,
			store       =  6,
			store_n     =  7,
			getstatic   = 11,
			putstatic   = 12,
			getfield    = 13,
			putfield    = 14,
			const_n     = 15,
			const_m1    = 21,
			const_      = 22,
			add         = 23,
			sub         = 24,
			mul         = 25,
			div         = 26,
			rem         = 27,
			neg         = 28,
			shl         = 29,
			shr         = 30,
			inc         = 31,
			new_        = 32,
			newarray    = 33,
			aload       = 34,
			astore      = 35,
			baload      = 36,
			bastore     = 37,
			arraylength = 38,
			pop         = 39,
			dup         = 40,
			dup2        = 41,
			jmp         = 42,
			jcc         = 43,
			call        = 49,
			return_     = 50,
			enter       = 51,
			exit        = 52,
			read        = 53,
			print       = 54,
			bread       = 55,
			bprint      = 56,
			trap        = 57;

	// error messages
	private static final String
		CODE_FULL  = ErrorStrings.CODE_FULL,
		CODE_LOAD  = ErrorStrings.CODE_LOAD,
		NO_VAR     = ErrorStrings.NO_VAR,
		NO_VAR_INC = ErrorStrings.NO_VAR_INC,
		OBJ_FILE   = ErrorStrings.OBJ_FILE;

	// codes for compare operators
	public static final int eq = 0, ne = 1, lt = 2, le = 3, gt = 4, ge = 5;

	private static final int [] inverse = { ne, eq, ge, gt, le, lt };

	/** Maximum size of generated code (in bytes). */
	private static final int maxCodeSize = 8192;
	
	/**	Code buffer */
	private static byte[] buf = new byte[maxCodeSize];
	
	/** True will code buffer is not full. */
	private static boolean generatingCode = true;

	/**	Program counter.
	 *	Indicates next free byte in code buffer. 
	 */
	public static int pc = 0;

	/**	PC of main method (set by parser). */
	public static int mainpc = -1;

	/**	Length of static data in words (set by parser). */
	public static int dataSize = 0;

	//----- initialization

	public static void init () {
		buf = new byte[maxCodeSize];
		generatingCode = true;
		pc = 0;
		mainpc = -1;
		dataSize = 0;
	}

	//----- code storage management

	public static void put (int x) { 
		if (generatingCode)
			if (pc == maxCodeSize) {
					generatingCode = false;
					Parser.Errors.semError(CODE_FULL);
			} else buf[pc++] = (byte) x; 
	}
	public static void put2 (int x) { put(x>>8); put(x); }
	public static void put4 (int x) { put2(x>>16); put2(x); }
	public static void put2 (int pos, int x) {
		int oldpc = pc; pc = pos;
		put2(x);
		pc = oldpc;
	}

	public static int get  (int pos) { return buf[pos]; }
	public static int get2 (int pos) { 
		return (get(pos) << 8) + (get(pos+1) & 0xFF); 
	}

	// ---------- instruction generation

	/**	Load the operand x onto the expression stack. */
	public static void load (Item x) {
		/* TODO: add your own code here */
	}

	/**	Load an integer constant onto the expression stack. */
	public static void loadConst (int n) {
		/* TODO: add your own code here */
	}

	/** Generate an assignment x = y. */
	public static void assign (Item x, Item y) {
		/* TODO: add your own code here */
	}

	/**	Generate an increment instruction that increments x by n. */
	public static void inc (Item x, int n) {
		/* TODO: add your own code here */
	}

	/** Unconditional jump. */
	public static void jump (Label lab) { 
		/* TODO: add your own code here */
	}

	/** True Jump.
	 *	Generates conditional jump instruction and links it to true jump chain.
	 */
	public static void tJump (Item x) {
		/* TODO: add your own code here */
	}

	/** False Jump.
	 *	Generates conditional jump instruction and links it to false jump chain.
	 */
	public static void fJump (Item x) {
		/* TODO: add your own code here */
	}

	/** Write the code buffer to the output stream. */
	public static void write (OutputStream os) {
		try {
			int codeSize = pc;
//			Decoder.decode(buf, 0, codeSize);  // uncomment for debugging output
			put('M'); put('J');
			put4(codeSize);
			put4(dataSize);
			put4(mainpc);
			os.write(buf, codeSize, pc - codeSize);  // header
			os.write(buf, 0, codeSize);              // code
			os.flush();
			os.close();
		} catch (IOException ioe) {
			Parser.Errors.semError(OBJ_FILE);
		}
	}
}
